from bullmq.types import BackoffOptions

import math


class Backoffs:

    builtin_strategies = {
        "fixed": lambda delay: lambda attempts_made, type, err, job: delay,
        "exponential": lambda delay: lambda attempts_made, type, err, job: int(round(pow(2, attempts_made - 1) * delay))
    }    

    @staticmethod
    def normalize(backoff: int | BackoffOptions):
        if type(backoff) == int and math.isfinite(backoff):
            return {
                "type": "fixed",
                "delay": backoff
            }
        elif backoff:
            return backoff

    @staticmethod
    async def calculate(backoff: BackoffOptions, attempts_made: int, err, job, customStrategy):
        if backoff:
            strategy = lookup_strategy(backoff, customStrategy)

            return strategy(attempts_made, backoff.get("type"), err, job)


def lookup_strategy(backoff: BackoffOptions, custom_strategy):
    backoff_type = backoff.get("type")
    if backoff_type in Backoffs.builtin_strategies:
        Backoffs.builtin_strategies[backoff.type](backoff.delay)
    elif custom_strategy:
        return custom_strategy
    else:
        raise Exception(f"Unknown backoff strategy {backoff_type}.If a custom backoff strategy is used, specify it when the queue is created.")



# Credits: https://gist.github.com/marc-x-andre/1c55b3fafd1d00cfdaa205ec53a08cf3
from typing import Dict


class EventEmitter:

    def __init__(self):
        self._callbacks: Dict[str, callable] = {}

    def on(self, event_name: str, function):
        self._callbacks[event_name] = self._callbacks.get(
            event_name, []) + [function]
        return function

    def emit(self, event_name: str, *args, **kwargs):
        [function(*args, **kwargs)
         for function in self._callbacks.get(event_name, [])]

    def off(self, event_name: str, function):
        self._callbacks.get(event_name, []).remove(function)



"""
BullMQ

A background job processor and message queue for Python based on Redis.
"""
__version__ = "1.14.0"
__author__ = 'Taskforce.sh Inc.'
__credits__ = 'Taskforce.sh Inc.'

from bullmq.queue import Queue
from bullmq.job import Job
from bullmq.flow_producer import FlowProducer
from bullmq.worker import Worker
from bullmq.custom_errors import WaitingChildrenError



import redis.asyncio as redis
from redis.backoff import ExponentialBackoff
from redis.retry import Retry
from redis.exceptions import (
   BusyLoadingError,
   ConnectionError,
   TimeoutError
)
import warnings

class RedisConnection:
    """
    RedisConnection class
    """

    minimum_version = '5.0.0'
    recommended_minimum_version = '6.2.0'

    def __init__(self, redisOpts: dict | str = {}):
        self.version = None
        retry = Retry(ExponentialBackoff(), 3)
        retry_errors = [BusyLoadingError, ConnectionError, TimeoutError]

        if isinstance(redisOpts, dict):
            host = redisOpts.get("host") or "localhost"
            port = redisOpts.get("port") or 6379
            db = redisOpts.get("db") or 0
            password = redisOpts.get("password") or None
            username = redisOpts.get("username") or None

            self.conn = redis.Redis(
                host=host, port=port, db=db, password=password, decode_responses=True,
                retry=retry, retry_on_error=retry_errors, username=username)
        else:
            self.conn = redis.from_url(redisOpts, decode_responses=True, retry=retry,
                retry_on_error=retry_errors)

    def disconnect(self):
        """
        Disconnect from Redis
        """
        return self.conn.disconnect()

    def close(self):
        """
        Close the connection
        """
        return self.conn.close()

    async def getRedisVersion(self):
        if self.version is not None:
            return self.version

        doc = await self.conn.info()
        if doc.get("maxmemory_policy") != "noeviction":
            warnings.warn(f'IMPORTANT! Eviction policy is {doc.get("maxmemory_policy")}. It should be "noeviction"')

        self.version = doc.get("redis_version")
        return doc.get("redis_version")



import semver
import traceback

def isRedisVersionLowerThan(current_version, minimum_version):
    return semver.compare(current_version, minimum_version) == -1

def extract_result(job_task):
    try:
        return job_task.result()
    except Exception as e:
        if not str(e).startswith('Connection closed by server'):
            # lets use a simple-but-effective error handling:
            # print error message and ignore the job
            print("ERROR:", e)
            traceback.print_exc()

def get_parent_key(opts: dict):
    if opts:
        return f"{opts.get('queue')}:{opts.get('id')}"



import asyncio

# Credits: https://stackoverflow.com/questions/45419723/python-timer-with-asyncio-coroutine


class Timer:
    def __init__(self, interval: int, callback, *args, **kwargs):
        self.interval = interval
        self.args = args
        self.kwargs = kwargs
        self.callback = callback
        self._ok = True
        self._task = asyncio.ensure_future(self._job())

    async def _job(self):
        try:
            while self._ok:
                await asyncio.sleep(self.interval)
                await self.callback(*self.args, **self.kwargs)
        except Exception as ex:
            print(ex)

    def stop(self):
        self._ok = False
        self._task.cancel()
